# -*- coding: utf-8 -*-

from elasticsearch import Elasticsearch
from elasticsearch_dsl import Search


def get_client():
    return Elasticsearch(['http://localhost:8200'], timeout=60)


def get_search(client=None, indices=None):
    indices = indices or '*'
    client = client or get_client()
    return Search(using=client, index=indices)



# -*- coding: utf-8 -*-

from datetime import timedelta, datetime

from fiqs.exceptions import MissingParameterException


class Metric(object):
    def is_range(self):
        return False

    def is_field_agg(self):
        return NotImplemented

    def is_doc_count(self):
        return False

    def is_computed(self):
        return False


class ModelMetric(Metric):
    def __init__(self, model):
        self.model = model

    def is_field_agg(self):
        return False


class Count(ModelMetric):
    def is_doc_count(self):
        return True

    def __str__(self):
        return 'doc_count'

    def order_by_key(self):
        return '_count'


class Aggregate(Metric):
    def reference(self):
        if hasattr(self, 'ref'):
            return self.ref
        return self.__class__.__name__.lower()

    def __init__(self, field, **kwargs):
        self.field = field
        self.params = kwargs

    def is_field_agg(self):
        return True

    def __str__(self):
        model = self.field.model.__name__.lower()
        op = self.__class__.__name__.lower()
        return '{}__{}__{}'.format(model, self.field.key, op)

    def agg_params(self):
        params = {
            'name': self.field.key,
            'field': self.field.get_storage_field(),
            'agg_type': self.reference(),
        }
        return params

    def choice_keys(self):
        return None

    def get_casted_value(self, v):
        return self.field.get_casted_value(v)


class Avg(Aggregate):
    def get_casted_value(self, v):
        """Average of an IntegerField does not have to be an integer"""
        return v


class Max(Aggregate): pass
class Min(Aggregate): pass
class Sum(Aggregate): pass
class Cardinality(Aggregate): pass


class Histogram(Aggregate):
    ref = 'histogram'

    def agg_params(self):
        params = super(Histogram, self).agg_params()
        params.update({
            'min_doc_count': 0,
        })

        if 'min' in self.params and 'max' in self.params:
            self.min = self.params.pop('min')
            self.max = self.params.pop('max')

            params['extended_bounds'] = {
                'min': self.min,
                'max': self.max,
            }

        params.update(self.params)

        if 'interval' not in params:
            raise MissingParameterException('missing interval parameter')

        self.interval = params['interval']

        return params


def get_timedelta_from_timestring(timestring):
    pass


TIME_UNIT_CONVERSION = {
    'd': 'days',
    'day': 'days',
    'H': 'hours',
    'h': 'hours',
    'hour': 'hours',
    'm': 'minutes',
    'minute': 'minutes',
    's': 'seconds',
    'second': 'seconds',
}


def get_timedelta_from_interval(interval):
    # Some intervals are still missing: year, quarter, month, week
    for key, param in TIME_UNIT_CONVERSION.items():
        if interval.endswith(key):
            value = interval.rstrip(key)
            if not value:
                value = '1'

            return timedelta(**{
                param: int(value),
            })

    return None


def get_rounded_date_from_timedelta(d, delta):
    epoch = datetime(1970, 1, 1)
    nb_seconds = int((d - epoch).total_seconds())
    delta_seconds = int(delta.total_seconds())

    rounded_nb_seconds = (nb_seconds // delta_seconds) * delta_seconds
    return datetime.utcfromtimestamp(rounded_nb_seconds)


class DateHistogram(Histogram):
    ref = 'date_histogram'

    def choice_keys(self):
        if not hasattr(self, 'min') or not hasattr(self, 'max'):
            return None

        delta = get_timedelta_from_interval(self.interval)
        if not delta:
            return None

        start = get_rounded_date_from_timedelta(self.min, delta)
        end = self.max

        choice_keys = []
        current = start
        while current <= end:
            choice_keys.append(current)
            current += delta

        return choice_keys


class DateRange(Aggregate):
    ref = 'date_range'

    def agg_params(self):
        params = super(DateRange, self).agg_params()
        params.update(self.params)

        if 'ranges' not in params:
            raise MissingParameterException('missing ranges parameter')

        return params


class Operation(Metric):
    def is_field_agg(self):
        return False

    def is_computed(self):
        return True

    def compute_one(self, row):
        raise NotImplementedError

    def compute(self, results, key=None):
        raise NotImplementedError

    def get_casted_value(self, v):
        return v

    @property
    def operands(self):
        return self._operands

    def __init__(self, *args):
        self._operands = args


def div_or_none(a, b, percentage=False):
    base = 100.0 if percentage else 1.0
    if b and a is not None:
        return base * a / b
    return None


def add_or_none(operands):
    if any([op is None for op in operands]):
        return None
    return sum(operands)


def sub_or_none(a, b):
    if a is not None and b is not None:
        return a - b
    return None


class Ratio(Operation):
    def __init__(self, dividend, divisor):
        super(Ratio, self).__init__(dividend, divisor)

        self.dividend = dividend
        self.divisor = divisor

    def __str__(self):
        return '{}__div__{}'.format(self.dividend, self.divisor)

    def compute_one(self, row):
        dividend = row[str(self.dividend)]

        if not self.divisor.is_computed():
            divisor = row[str(self.divisor)]
        else:
            divisor = self.divisor.compute_one(row)

        return div_or_none(dividend, divisor, percentage=True)


class Addition(Operation):
    def __str__(self):
        return '__add__'.join(['{}'.format(op) for op in self.operands])

    def compute_one(self, row):
        keys = [str(op) for op in self.operands]
        return add_or_none([row[key] for key in keys])


class Subtraction(Operation):
    def __init__(self, minuend, subtraend):
        super(Subtraction, self).__init__(minuend, subtraend)

        self.minuend = minuend
        self.subtraend = subtraend

    def __str__(self):
        return '{}__sub__{}'.format(self.minuend, self.subtraend)

    def compute_one(self, row):
        key_a = str(self.minuend)
        key_b = str(self.subtraend)

        return sub_or_none(row[key_a], row[key_b])



# -*- coding: utf-8 -*-
#!/usr/bin/env python

import sys

import pytest


def run_all(argv=None):
    # always insert coverage when running tests through setup.py
    if argv is None:
        argv = [
            '--cov', 'fiqs', '--verbose',
            '--junitxml', 'junit.xml', '--cov-report', 'xml',
        ]
    else:
        argv = argv[1:]

    sys.exit(pytest.main(['fiqs'] + argv))


if __name__ == '__main__':
    run_all(sys.argv)



# -*- coding: utf-8 -*-

import copy
from datetime import datetime

from fiqs.i18n import _


class Field(object):
    def __init__(self, type, key=None, verbose_name=None, storage_field=None,
                 unit=None, choices=None, data=None, parent=None):
        verbose_name = verbose_name or key
        storage_field = storage_field or key
        choices = choices or ()
        data = data or {}

        self.key = key
        self.type = type
        self.verbose_name = verbose_name
        self.storage_field = storage_field
        self.unit = unit
        self.choices = choices
        self.data = data
        self.parent = parent

    def get_copy(self):
        return self.__class__(
            key=self.key, verbose_name=self.verbose_name, storage_field=self.storage_field,
            unit=self.unit, choices=self.choices, data=self.data, parent=self.parent,
        )

    def __repr__(self):
        if hasattr(self, 'model'):
            return '<{}: {}.{}>'.format(
                self.__class__.__name__,
                self.model.__name__,
                self.key,
            )

        return '<{}: {}>'.format(
                self.__class__.__name__,
                self.key,
        )

    def _set_key(self, key):
        self.key = key
        if not self.verbose_name:
            self.verbose_name = key
        if not self.storage_field:
            self.storage_field = key

    def get_storage_field(self):
        parent_field = self.get_parent_field()
        if not parent_field:
            return self.storage_field

        return '{}.{}'.format(
            parent_field.get_storage_field(),
            self.storage_field,
        )

    def bucket_params(self):
        d = {
            'name': self.key,
            'field': self.get_storage_field(),
        }

        if 'script' in self.data:
            # should we remove field?
            d['script'] = self.data['script'].format('_value')

        if 'size' in self.data:
            size = self.data['size']
            if size == 0:
                size = 2 ** 31 - 1
            d['size'] = size

        return d

    def is_range(self):
        return 'ranges' in self.data

    def _has_pretty_choices(self):
        return self.choices and isinstance(self.choices[0], tuple)

    def choice_keys(self):
        if self._has_pretty_choices():
            return [choice[0] for choice in self.choices]

        if self.choices:
            return self.choices

        if self.is_range():
            return [r['key'] for r in self.data['ranges']]

        raise NotImplementedError()

    def min_key(self):
        if self.data and 'min' in self.data:
            return self.data['min']

        if self.choices:
            return min(self.choice_keys())

    def max_key(self):
        if self.data and 'max' in self.data:
            return self.data['max']

        if self.choices:
            return max(self.choice_keys())

    def get_parent_field(self):
        if not self.parent:
            return None
        return getattr(self.model, self.parent)

    def get_casted_value(self, v):
        return v


class TextField(Field):
    def __init__(self, **kwargs):
        super(TextField, self).__init__('text', **kwargs)


class KeywordField(Field):
    def __init__(self, **kwargs):
        super(KeywordField, self).__init__('keyword', **kwargs)


class DateField(Field):
    def __init__(self, **kwargs):
        super(DateField, self).__init__('date', **kwargs)

    def get_casted_value(self, v):
        # Careful, we lose the milliseconds here
        return datetime.utcfromtimestamp(v / 1000)


class BaseIntegerField(Field):
    def __init__(self, **kwargs):
        super(BaseIntegerField, self).__init__(self.type, **kwargs)

    def get_casted_value(self, v):
        return int(v) if v else v


class LongField(BaseIntegerField):
    type = 'long'


class IntegerField(BaseIntegerField):
    type = 'integer'


class ShortField(BaseIntegerField):
    type = 'short'


class ByteField(BaseIntegerField):
    type = 'byte'


class BaseFloatField(Field):
    def __init__(self, **kwargs):
        super(BaseFloatField, self).__init__(self.type, **kwargs)

    def get_casted_value(self, v):
        return float(v) if v else v


class DoubleField(BaseFloatField):
    type = 'double'


class FloatField(BaseFloatField):
    type = 'float'


def get_weekdays():
    return [
        (0, _('Monday')),
        (1, _('Tuesday')),
        (2, _('Wednesday')),
        (3, _('Thursday')),
        (4, _('Friday')),
        (5, _('Saturday')),
        (6, _('Sunday')),
    ]


def get_iso_weekdays():
    return [
        (1, _('Monday')),
        (2, _('Tuesday')),
        (3, _('Wednesday')),
        (4, _('Thursday')),
        (5, _('Friday')),
        (6, _('Saturday')),
        (7, _('Sunday')),
    ]


class DayOfWeekField(ByteField):
    def __init__(self, iso=True, **kwargs):
        if iso:
            choices = get_iso_weekdays()
            data = {'min': 1, 'max': 7}
        else:
            choices = get_weekdays()
            data = {'min': 0, 'max': 6}

        kwargs['choices'] = choices
        kwargs['data'] = data

        super(DayOfWeekField, self).__init__(**kwargs)


class HourOfDayField(ByteField):
    def __init__(self, **kwargs):
        kwargs['choices'] = [(i, _('{hour}h').format(hour=i)) for i in range(24)]
        kwargs['data'] = {'min': 0, 'max': 23}

        super(HourOfDayField, self).__init__(**kwargs)


class BooleanField(Field):
    def __init__(self, **kwargs):
        super(BooleanField, self).__init__('boolean', **kwargs)


class NestedField(Field):
    def __init__(self, **kwargs):
        super(NestedField, self).__init__('nested', **kwargs)

    def nested_params(self):
        params = {
            'name': self.key,
            'agg_type': 'nested',
            'path': self.get_storage_field(),
        }
        return params


class ReverseNestedField(Field):
    def __init__(self, **kwargs):
        super(ReverseNestedField, self).__init__('reverse_nested', **kwargs)

    def nested_params(self):
        params = {
            'name': self.key,
            'agg_type': 'reverse_nested',
        }
        # /!\ path must not be provided for root reverse_nested aggregation
        if self.key != 'root':
            params['path'] = self.key

        return params


class FieldWithChoices(Field):
    def __init__(self, field, choices=None):
        choices = choices or ()
        data = copy.deepcopy(field.data)
        return super(FieldWithChoices, self).__init__(
            field.type,
            key=field.key,
            verbose_name=field.verbose_name,
            storage_field=field.storage_field,
            unit=field.unit,
            choices=choices,
            data=data,
            parent=field.parent,
        )


class DataExtendedField(Field):
    def __init__(self, field, **kwargs):
        data = copy.deepcopy(field.data)
        data.update(**kwargs)
        return super(DataExtendedField, self).__init__(
            field.type,
            key=field.key,
            verbose_name=field.verbose_name,
            storage_field=field.storage_field,
            unit=field.unit,
            choices=field.choices,
            data=data,
            parent=field.parent,
        )



# -*- coding: utf-8 -*-

import gettext
import os
import sys


if sys.version_info[0] >= 3:
    _ = gettext.gettext
else:
    localedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'locale')
    translate = gettext.translation('fiqs', localedir, fallback=True)
    _ = translate.ugettext



# -*- coding: utf-8 -*-

import copy

import pytest

from fiqs import fields
from fiqs.exceptions import FieldError
from fiqs.models import Model
from fiqs.testing.models import Sale, SaleWithProducts, SaleWithParts, SaleWithoutProducts


EXPECTED_MAPPING = {
        'sale': {
            'dynamic': 'strict',
            'properties': {
                'id': {'type': 'integer'},
                'shop_id': {'type': 'integer'},
                'client_id': {'type': 'keyword'},
                'timestamp': {'type': 'date'},
                'price': {'type': 'integer'},
                'payment_type': {'type': 'keyword'},
                'products': {
                    'properties': {
                        'product_id': {'type': 'keyword'},
                        'product_price': {'type': 'integer'},
                        'product_type': {'type': 'keyword'},
                        'parts': {
                            'properties': {
                                'part_id': {'type': 'keyword'},
                                'warehouse_id': {'type': 'keyword'},
                                'part_price': {'type': 'integer'},
                            },
                            'type': 'nested',
                        },
                    },
                    'type': 'nested',
                },
            },
        },
    }


def test_mapping_from_model():
    assert Sale.get_mapping().to_dict() == EXPECTED_MAPPING


def test_mapping_from_model_child_class():
    mapping_without_parts = copy.deepcopy(EXPECTED_MAPPING)
    mapping_without_parts['sale']['properties']['products']['properties'].pop('parts')
    mapping_without_parts['sale_with_products'] = mapping_without_parts.pop('sale')
    assert SaleWithProducts.get_mapping().to_dict() == mapping_without_parts

    mapping_with_parts = copy.deepcopy(EXPECTED_MAPPING)
    mapping_with_parts['sale_with_parts'] = mapping_with_parts.pop('sale')
    assert SaleWithParts.get_mapping().to_dict() == mapping_with_parts


def test_child_class_redefines_parents_field():
    with pytest.raises(FieldError):
        class Parent(Model):
            field = fields.IntegerField()

        class Child(Parent):
            field = fields.IntegerField()


def test_field_model_access_child():
    assert SaleWithoutProducts.price.model == SaleWithoutProducts
    assert SaleWithProducts.price.model == SaleWithProducts
    assert SaleWithParts.price.model == SaleWithParts



# -*- coding: utf-8 -*-

from os.path import join, dirname
from setuptools import setup, find_packages
from setuptools.command.sdist import sdist


VERSION = (0, 3, 10)
__version__ = VERSION
__versionstr__ = '.'.join(map(str, VERSION))


f = open(join(dirname(__file__), 'README.md'))
long_description = f.read().strip()
f.close()


install_requires = [
    'urllib3>=1.8, <2.0',
    'elasticsearch>=5.1.0',
    'elasticsearch-dsl>=5.1.0',
]
setup_requires = [
    'Babel>=2.3.4',
]
tests_require = [
    'six==1.10.0',
    'Faker==0.7.3',
    'pytest==3.0.5',
    'pytest-cov',
]


class Sdist(sdist):
    """Custom ``sdist`` command to ensure that mo files are always created."""

    def run(self):
        self.run_command('compile_catalog')
        # sdist is an old style class so super cannot be used.
        sdist.run(self)


setup(
    name='fiqs',
    description="Python client for Elasticsearch built on top of elasticsearch-dsl",
    license="MIT License",
    url="https://gitlab.com/pmourlanne/fiqs",
    long_description=long_description,
    version=__versionstr__,
    author="Pierre Mourlanne",
    author_email="pmourlanne@gmail.com",
    packages=find_packages(
        where='.',
    ),
    install_requires=install_requires,
    setup_requires=setup_requires,
    cmdclass={'sdist': Sdist},
    test_suite='fiqs.tests.run_tests.run_all',
    tests_require=tests_require,
)
